@@ -14,5 +14,5 @@ in conjunction with your Project to filter your data.
 
 from .project import Project
 from .query import Query, QueryGroup
-from .request import RCRequest, RCAPIError
+from .request import RCRequest, RCAPIError, RedcapError
 from .version import VERSION as __version__
\ No newline at end of file
