@@ -14,12 +14,13 @@ from .request import RCRequest
 class Project(object):
     """Main class representing a RedCap Project"""
 
-    def __init__(self, url, token, name=''):
+    def __init__(self, url, token, name='', verify_ssl=True):
         """Must init with your token"""
 
         self.token = token
         self.name = name
         self.url = url
+        self.verify = verify_ssl
 
         self.metadata = self.__md()
         self.field_names = self.filter_metadata('field_name')
@@ -28,16 +29,14 @@ class Project(object):
         self.field_labels = self.filter_metadata('field_label')
         self.forms = tuple(set(c['form_name'] for c in self.metadata))
         # determine whether longitudinal
-        ev_data = RCRequest(self.url, self.__basepl('event'),
-            'exp_event').execute()[0]
+        ev_data = self._call_api(self.__basepl('event'), 'exp_event')[0]
         if 'error' in ev_data:
             events = tuple([])
             arm_nums = tuple([])
             arm_names = tuple([])
         else:
             events = ev_data
-            arm_data = RCRequest(self.url, self.__basepl('arm'),
-                'exp_arm').execute()[0]
+            arm_data = self._call_api(self.__basepl('arm'), 'exp_arm')[0]
             arm_nums = tuple([a['arm_num'] for a in arm_data])
             arm_names = tuple([a['name'] for a in arm_data])
         self.events = events
@@ -48,16 +47,22 @@ class Project(object):
         """Return the project's metadata structure"""
         p_l = self.__basepl('metadata')
         p_l['content'] = 'metadata'
-        return RCRequest(self.url, p_l, 'metadata').execute()[0]
+        return self._call_api(p_l, 'metadata')[0]
 
     def __basepl(self, content, rec_type='flat', format='json'):
         """Return a dictionary which can be used as is or added to for
-        RCRequest payloads"""
+        payloads"""
         d = {'token': self.token, 'content': content, 'format': format}
         if content not in ['metadata', 'file']:
             d['type'] = rec_type
         return d
 
+    def is_longitudinal(self):
+        """Return boolean whether this project is longitudinal or not"""
+        return len(self.events) > 0 and \
+            len(self.arm_nums) > 0 and \
+            len(self.arm_names) > 0
+
     def filter_metadata(self, key):
         """Return a list values for the key in each field from the project's
         metadata.
@@ -72,8 +77,56 @@ class Project(object):
             raise KeyError("Key not found in metadata")
         return filtered
 
+    def _kwargs(self):
+        """Private method to build a dict for sending to RCRequest
+
+        Other default kwargs to the http library should go here"""
+        return {'verify': self.verify}
+
+    def _call_api(self, payload, typpe, **kwargs):
+        request_kwargs = self._kwargs()
+        request_kwargs.update(kwargs)
+        rcr = RCRequest(self.url, payload, typpe)
+        return rcr.execute(**request_kwargs)
+
+    def export_fem(self, arms=None, format='obj', df_kwargs=None):
+        """Export the project's form to event mapping
+
+        Parameters
+        ----------
+        arms: list
+            Limit exported form event mappings to these arm numbers
+        format: {'obj', 'csv', 'xml'} default obj
+            Return the form event mappings in native objects, 
+            csv or xml, df will return a pandas.DataFrame
+        df_kwargs: dict
+            Passed to pandas.read_csv to control construction of
+            returned DataFrame
+        """
+        ret_format = format
+        if format == 'obj':
+            ret_format = 'json'
+        if format == 'df':
+            from StringIO import StringIO
+            from pandas import read_csv
+            ret_format = 'csv'
+        pl = self.__basepl('formEventMapping', format=ret_format)
+        to_add = [arms]
+        str_add = ['arms']
+        for key, data in zip(str_add, to_add):
+            if data:
+                pl[key] = ','.join(data)
+        response, _ = self._call_api(pl, 'exp_fem')
+        if format in ('obj', 'csv', 'xml'):
+            return response
+        elif format == 'df':
+            if not df_kwargs:
+                return read_csv(StringIO(response))
+            else:
+                return read_csv(StringIO(response), **df_kwargs)
+
     def export_metadata(self, fields=None, forms=None, format='obj',
-            df_kwargs=None):
+                        df_kwargs=None):
         """Export the project's metadata
 
         Parameters
@@ -102,7 +155,7 @@ class Project(object):
         for key, data in zip(str_add, to_add):
             if data:
                 pl[key] = ','.join(data)
-        response = RCRequest(self.url, pl, 'metadata').execute()[0]
+        response, _ = self._call_api(pl, 'metadata')
         if format in ('obj', 'csv', 'xml'):
             return response
         elif format == 'df':
@@ -111,8 +164,8 @@ class Project(object):
             return read_csv(StringIO(response), **df_kwargs)
 
     def export_records(self, records=None, fields=None, forms=None,
-                events=None, raw_or_label='raw', event_name='label',
-                format='obj', df_kwargs=None):
+                       events=None, raw_or_label='raw', event_name='label',
+                       format='obj', df_kwargs=None):
         """Return data
 
         High level function of Project
@@ -154,9 +207,9 @@ class Project(object):
             ret_format = 'csv'
         pl = self.__basepl('record', format=ret_format)
         keys_to_add = (records, fields, forms, events,
-                        raw_or_label, event_name)
+                       raw_or_label, event_name)
         str_keys = ('records', 'fields', 'forms', 'events', 'rawOrLabel',
-                'eventName')
+                    'eventName')
         for key, data in zip(str_keys, keys_to_add):
             if data:
                 #  Make a url-ok string
@@ -164,25 +217,32 @@ class Project(object):
                     pl[key] = ','.join(data)
                 else:
                     pl[key] = data
-        response = RCRequest(self.url, pl, 'exp_record').execute()[0]
+        response, _ = self._call_api(pl, 'exp_record')
         if format in ('obj', 'csv', 'xml'):
             return response
         elif format == 'df':
             if not df_kwargs:
-                df_kwargs = {'index_col': self.def_field}
-            return read_csv(StringIO(response), **df_kwargs)
+                if self.is_longitudinal():
+                    df_kwargs = {'index_col': [self.def_field,
+                                               'redcap_event_name']}
+                else:
+                    df_kwargs = {'index_col': self.def_field}
+            buf = StringIO(response)
+            df = read_csv(buf, **df_kwargs)
+            buf.close()
+            return df
 
     def metadata_type(self, field_name):
         """If the given field_name is validated by REDCap, return it's type"""
         return self.__meta_metadata(field_name,
-            'text_validation_type_or_show_slider_number')
+                                    'text_validation_type_or_show_slider_number')
 
     def __meta_metadata(self, field, key):
         """Return the value for key for the field in the metadata"""
         mf = ''
         try:
             mf = str([f[key] for f in self.metadata
-                            if f['field_name'] == field][0])
+                     if f['field_name'] == field][0])
         except IndexError:
             print("%s not in metadata field:%s" % (key, field))
             return mf
@@ -234,26 +294,58 @@ class Project(object):
                 print('%s --> %s' % (str(name), str(label)))
         return self.field_names, self.field_labels
 
-    def import_records(self, to_import, overwrite='normal'):
+    def import_records(self, to_import, overwrite='normal', format='json',
+                       return_format='json', return_content='count'):
         """ Import data into the RedCap Project
 
         Parameters
         ----------
-        to_import: seq of dicts
-            List of dictionaries describing the data you wish to import_records
+        to_import: seq of dicts | csv/xml string | pandas.DataFrame
+            List of dictionaries describing the data you wish to
+            import_records, csv/xml string represenation of your data or
+            a pandas.DataFrame
             Note:
                 Keys of the dictionaries should be subset of project's,
                 fields, but this isn't a requirement. If you provide keys
                 that aren't defined fields, the returned response will
-                contain and 'error' key.
-        overwrite: 'normal' | 'overwrite'
+                contain an 'error' key.
+        overwrite: ('normal') | 'overwrite'
             'overwrite' will erase values previously stored in the database if
             not specified in the to_import dictionaries
+        format: ('json') | 'xml' | 'csv'
+            Format of incoming data. By default, to_import will be json-encoded
+        return_format: ('json') | 'csv' | 'xml'
+            Response format. By default, response will be json-decoded.
+        return_content: ('count') | 'ids' | 'nothing'
+            By default, the response contains a 'count' key with the number of
+                records just imported. By specifying 'ids', a list of ids
+                imported will be returned. 'nothing' will only return
+                the HTTP status code and no message.
         """
         pl = self.__basepl('record')
+        if hasattr(to_import, 'to_csv'):
+            # We'll assume it's a df
+            from StringIO import StringIO
+            buf = StringIO()
+            if self.is_longitudinal():
+                csv_kwargs = {'index_label': [self.def_field,
+                                              'redcap_event_name']}
+            else:
+                csv_kwargs = {'index_label': self.def_field}
+            to_import.to_csv(buf, **csv_kwargs)
+            pl['data'] = buf.getvalue()
+            buf.close()
+            format = 'csv'
+        elif format == 'json':
+            pl['data'] = json.dumps(to_import, separators=(',', ':'))
+        else:
+            # don't do anything to csv/xml
+            pl['data'] = to_import
         pl['overwriteBehavior'] = overwrite
-        pl['data'] = json.dumps(to_import, separators=(',', ':'))
-        return RCRequest(self.url, pl, 'imp_record').execute()[0]
+        pl['format'] = format
+        pl['returnFormat'] = return_format
+        pl['returnContent'] = return_content
+        return self._call_api(pl, 'imp_record')[0]
 
     def export_file(self, record, field, event=None, return_format='json'):
         """ Export the contents of a file stored for a particular record
@@ -283,19 +375,19 @@ class Project(object):
         pl['record'] = record
         if event:
             pl['event'] = event
-        content, headers = RCRequest(self.url, pl, 'exp_file').execute()
+        content, headers = self._call_api(pl, 'exp_file')
         #REDCap adds some useful things in content-type
         if 'content-type' in headers:
             splat = [kv.strip() for kv in headers['content-type'].split(';')]
             kv = [(kv.split('=')[0], kv.split('=')[1].replace('"', '')) for kv
-                in splat if '=' in kv]
+                  in splat if '=' in kv]
             content_map = dict(kv)
         else:
             content_map = {}
         return content, content_map
 
     def import_file(self, record, field, fname, fobj, event=None,
-            return_format='json'):
+                    return_format='json'):
         """Import the contents of a file represented by fobj to a
         particular records field
 
@@ -325,7 +417,7 @@ class Project(object):
         if event:
             pl['event'] = event
         file_kwargs = {'files': {'file': (fname, fobj)}}
-        return RCRequest(self.url, pl, 'imp_file').execute(**file_kwargs)[0]
+        return self._call_api(pl, 'imp_file', **file_kwargs)[0]
 
     def delete_file(self, record, field, return_format='json', event=None):
         self._check_file_field(field)
@@ -338,7 +430,7 @@ class Project(object):
         pl['field'] = field
         if event:
             pl['event'] = event
-        return RCRequest(self.url, pl, 'del_file').execute()[0]
+        return self._call_api(pl, 'del_file')[0]
 
     def _check_file_field(self, field):
         """Check that field exists and is a file field"""
@@ -377,4 +469,4 @@ class Project(object):
         list of users dicts when format=json, otherwise a string
         """
         pl = self.__basepl(content='user', format=format)
-        return RCRequest(self.url, pl, 'exp_user').execute()[0]
+        return self._call_api(pl, 'exp_user')[0]
